--local StatLogic = LibStub:GetLibrary("LibStatLogic-1.1")
local StatLogic = LibStub("LibStatLogic-1.2")
local Toolkit = LibStub:GetLibrary("Toolkit-1.0");

HealPointsCalculator ={
	HP_LastPowerfulCurrent = nil;
	--HP_LastPowerfulNew = nil;
	HP_LastEfficientCurrent = nil;
--	HP_LastEfficientNew = nil;
};

--[[
	function GetHPTick	
--]]

local BonusPercentInt;
local BonusPercentSpi;

-- Helper functions
--[[
	Returns the global cooldown time (normally 1.5 seconds) after
	it has been reduced by specifid amount of spell haste
--]]
local function getGCD(spellHaste)
	if (spellHaste == nil) then
		return 1.5;
	else
		return math.max(1.0, 1.5 / (1 + spellHaste / 100));
	end
end

--[[
	Get the amount of healing done by a particular spell
	@param spell A HealPointsSpells.SPELLTABLE entry the indicates the spell to be calculated
	@param healing A bonus amount of healing to be added to the spell
	@param spellCrit The healing crit chance
--]]
local function getHPCast(spell, healing, spellCrit)

	assert(spell, "calculator.lua:getHPCast - spell is nil");
	assert(healing, "calculator.lua:getHPCast - healing is nil");
	assert(spellCrit, "calculator.lua:getHPCast - spellCrit is nil");

	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactor, healFactor2 = HealPointsSpells:getHealingFactor(spell);
	spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);

	local out, minHeal, maxHeal;

	if (spell.class['type'] == 'normal' or spell.class['type'] == 'instant') then
		minHeal = (spell['min'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		maxHeal = (spell['max'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);

		--out = (((spell['min'] + spell['max']) / 2) + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		out = (minHeal+maxHeal) / 2;

	elseif (spell.class['type'] == 'hot') then
		local healTick = spell['tick'] + healing * healFactor / spell.class['tickcount']; -- +healing divided between original tickcount
		out = healTick * HealPointsSpells:getNumberOfTicks(spell) * percentBonus;

		minHeal = out;
		maxHeal = out;

	elseif (spell.class['type'] == 'regrowth') then

		-- Burst part
		minHeal = (spell['min'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		maxHeal = (spell['max'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);

		--out = (((spell['min'] + spell['max']) / 2) + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		out = (minHeal+maxHeal) / 2;

		-- HOT part
		local healTick = spell['tick'] + healing * healFactor2 / spell.class['tickcount'];
		out     = out     + healTick * HealPointsSpells:getNumberOfTicks(spell) * percentBonus;
		minHeal = minHeal + healTick * HealPointsSpells:getNumberOfTicks(spell) * percentBonus;
		maxHeal = minHeal + healTick * HealPointsSpells:getNumberOfTicks(spell) * percentBonus;

	elseif (spell.class['type'] == 'lifebloom') then
		-- HOT part
		local healTick = 
				spell['tick']
				--+ (healing + 2.777 * HealPointsBS:GetBonus('AVG_HOT_LIFEBL')) * healFactor / spell.class['tickcount'];
		out = healTick * HealPointsSpells:getNumberOfTicks(spell) * percentBonus;
		-- Burst part
		out = out + (
				spell['burst']
				--+ (healing + HealPointsBS:GetBonus('AVG_BURST_LIFEBL')) * healFactor2
				) * percentBonus * (1 + spellCrit / 200);

	elseif (spell.class['type'] == 'chain') then
		minHeal = (spell['min'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		maxHeal = (spell['max'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		--out = (((spell['min'] + spell['max']) / 2) + healing * healFactor) * percentBonus * (1 + spellCrit / 200);

		--local multiTargetFactor = 2.2; --(1+0.7+0.4+0.1);
		--local multiTargetFactor = 2.533; --assume four targets (1+0.7+0.4+0.1)
		--local multiTargetFactor = 2.1; --assume three targets (1+0.7+0.4)
		local multiTargetFactor = 1.7; --assume two targets only (1+0.7)

		minHeal = minHeal * multiTargetFactor; -- Assumes that the spell hits all 3 targets
		maxHeal = maxHeal * multiTargetFactor; -- Assumes that the spell hits all 3 targets

		--out = out * 1.75; -- Assumes that the spell hits all 3 targets
		out = (minHeal + maxHeal) / 2;

	elseif (spell.class['type'] == 'aoe') then
		out = (((spell['min'] + spell['max']) / 2) + healing * healFactor) * percentBonus * (1 + spellCrit / 200)
		out = out * math.min(5, HealPoints.db.char.hot['numtargets']);
	end

	--If Divine Plea is used it decreases our healing by 50% while it's active
	if (_G["HealPointsOptionsConfigDivinePlea"]:GetChecked() == 1) then -- Divine plea
		-- 50% reduction 9/120 of the time = 3.75% reduction on average
		out = out * 0.9625;
	end

  	return out, minHeal, maxHeal;
end

--[[
	Get HP healed per tick of a HOT spell
--]]
function HealPointsCalculator:GetHPTick(spell, healing)
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactor, healFactor2 = HealPointsSpells:getHealingFactor(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);

	local healTick = 0;

	if (spell.class['type'] == 'hot') then
		healTick = spell['tick'] + healing * healFactor / spell.class['tickcount']; -- +healing divided between original tickcount

	elseif (spell.class['type'] == 'regrowth') then
		healTick = spell['tick'] + healing * healFactor2 / spell.class['tickcount'];

	elseif (spell.class['type'] == 'lifebloom') then
		healTick = 
				spell['tick']
				--+ (healing + 2.777 * HealPointsBS:GetBonus('AVG_HOT_LIFEBL')) * healFactor / spell.class['tickcount'];
	end

	healTick = healTick * percentBonus;
	return healTick;
end

--[[
	Calculate the amount of healing per second done by a spell
--]]
local function getHPSec(spell, healing, spellCrit, spellHaste)
	if (spell.class['type'] == 'normal' or spell.class['type'] == 'chain' or spell.class['type'] == 'aoe') then
		local castTime = HealPointsSpells:GetCastTime(spell, spellHaste);
		if (castTime == 0) then
			castTime = getGCD(spellHaste);
		end
		if (spell.class['cooldown'] ~= nil) then
			castTime = math.max(castTime, spell.class['cooldown']);
		end
		return getHPCast(spell, healing, spellCrit) / castTime;
	elseif (spell.class['type'] == 'instant') then
		return getHPCast(spell, healing, spellCrit) / spell.class['cooldown'];
	elseif (spell.class['type'] == 'hot' or spell.class['type'] == 'lifebloom') then
		return getHPCast(spell, healing, spellCrit) / (HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval']);
	elseif (spell.class['type'] == 'regrowth') then
		return getHPCast(spell, healing, spellCrit) / (HealPointsSpells:GetCastTime(spell, spellHaste) + HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval']);
	end
end

--[[
	Calcuate the amount of healing done per point of mana by a spell
--]]
local function getHPMana(spell, healing, spellCrit)
	local hpcast = getHPCast(spell, healing, spellCrit);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	return hpcast / manaCost;
end

local function getHPTotalRegen(spell, healing, spellCrit, spellHaste, manaRegenCasting, mana)
	local hpPerCast = getHPCast(spell, healing, spellCrit);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);

	local interval; --how often the spell can be cast
	if (spell.class['type'] == 'regrowth') then
		interval = HealPointsSpells:GetCastTime(spell, spellHaste) + HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	elseif (spell.class['type'] == 'hot' or spell.class['type'] == 'lifebloom') then
		interval = HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	elseif (spell.class['type'] == 'instant') then
		interval = math.max(getGCD(spellHaste), spell.class['cooldown']);
	else
		interval = math.max(getGCD(spellHaste), HealPointsSpells:GetCastTime(spell, spellHaste)); 
	end

	local manaUsageSpell = manaCost - interval * manaRegenCasting;
	local hpPerMana = hpPerCast / manaUsageSpell;
	local hpPerManaBar = hpPerMana * mana;
	
	local numberCasts = mana / manaUsageSpell;
	local totalTime = numberCasts * interval;

	return hpPerManaBar, totalTime;
end

--[[
	Calculate the total amount of healing that can be done by a spell in the specified amount of time, starting at the specified mana level
	@param spell - A SPELLTABLE entry of the spell to calculate
	@param healing - A number that is an amount of bonus healing
	@param seconds - How long we run the calculation for
	@param startMana - The player's starting mana
	@param spellCrit - The player's chance of getting a critical heal
	@param spellHaste - The player's spell haste
	@param manaRegenCasting - The players mana regeneration per second while in combat
	@param manaRegenNormal - The players mana regeneration per second while out of combat
--]]
local function computeTotalHealingNormal(spell, healing, seconds, startMana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal)
	--DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell.class['name'].."("..spell['rank']..") secs="..seconds.." mana="..startMana,1,1,1);

	assert(spell);
	assert(healing);
	assert(seconds);
	assert(startMana);
	assert(spellCrit);
	assert(spellHaste);
	assert(manaRegenCasting);
	assert(manaRegenNormal);

	--simple bailout. If they configured the time for zero seconds then we couldn't heal anything
	if (seconds == 0) then
		return 0;
	end;

	--record the details of our calculations
	local details = {};
		--hpCast Health Points healed by a cast of the spell
		--castDelay Time the player must wait between casts (taking into effect cast time, cooldown, haste and gcd)
		--manaCost The effective mana cost of the spell (some players can have talents to refund mana, reducing effective mana cost)
		--hps Healing per second. Literally hpCast/castDelay. This number only makes sense while the player has mana to power through chain-cast
		--manaUsageSecond Mana used per second while chain-casting (takes into account that you can regen while casting)
		--secondsPhase1 Number of seconds the player can chain-cast before they run out of mana
		--totalHealedPhase1 The total amount of health healed in phase 1 chain-casting
		--secondsPhase2 Number of seconds that the player has to limp along waiting for mana to regen each cast, before time is up
		--cycleTime Number of seconds the player will have to wait between casts, as they wait for mana to regen
		--spellsCast
		--totalHealedPhase2 The total amount of health healed in phase 2 OOM casting


	local hpCast = getHPCast(spell, healing, spellCrit);
	local castDelay = math.max(getGCD(spellHaste), HealPointsSpells:GetCastTime(spell, spellHaste)); 

	if (spell.class['cooldown'] ~= nil) then
		castDelay = math.max(castDelay, spell.class['cooldown']);
	end

	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	local hps = hpCast / castDelay;

	--HealPoints:Print("computeTotalHealingNormal castDelay="..castDelay..", spell="..spell.class.abbr..", manaCost="..manaCost..", hpCast="..hpCast);

	-- Phase 1: Has mana to chain-cast
	local manaUsageSecond = (manaCost / castDelay) - manaRegenCasting;

	--HealPoints:Print("manaRegenCasting="..manaRegenCasting..", manaUsageSecond="..manaUsageSecond);

	--HealPoints:Print(string.format("[%s] (%.2f seconds) startMana=%.2f, manaUsageSecond=%.2f, manaCost=%.2f, castDelay=%.2f, manaRegenCasting=%.2f",
			--spell.class.abbr, seconds, startMana, manaUsageSecond, manaCost, castDelay, manaRegenCasting);

	local secondsPhase1 = math.min(seconds, startMana / manaUsageSecond);
	local totalHealed = secondsPhase1 * hps;

--	HealPoints:Print(string.format("[%s] Phase 1 (%s s): totalHealed=%.2f, secondsPhase1=%.2f, hps=%.2f", 
			--spell.class.abbr, secondsPhase1, totalHealed, hps));



	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
		-- Seconds between each casting
		local cycleTime = math.max(castDelay, manaCost / manaRegenCasting);
		local spellsCast = secondsPhase2 / cycleTime;
		local healedPhase2 = spellsCast * hpCast;
		totalHealed = totalHealed + healedPhase2;

		--HealPoints:Print(string.format("Phase 2: %s: totalHealed2=%.2f, secondsPhase2=%.2f, cycleTime=%.2f",
				--spell.class.abbr, healedPhase2, secondsPhase2, cycleTime));
	end

	return totalHealed;
end

local function computeTotalHealingRegrowth(spell, healing, seconds, startMana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal)
	assert(spell);
	assert(healing);
	assert(seconds);
	assert(startMana);
	assert(spellCrit);
	assert(spellHaste);
	assert(manaRegenCasting);
	assert(manaRegenNormal);

	--DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell.class['name'].."("..spell['rank']..") secs="..seconds.." mana="..startMana,1,1,1);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	local castTime = math.max(getGCD(spellHaste), HealPointsSpells:GetCastTime(spell, spellHaste)); 
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactorBurst, healFactorHot = HealPointsSpells:getHealingFactor(spell);
	local numtargets = HealPoints.db.char.hot['numtargets'];
	spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);

	local avgBurst = (((spell['min'] + spell['max']) / 2) + healing * healFactorBurst) * percentBonus * (1 + spellCrit / 200);
	local healTick = (spell['tick'] + healing * healFactorHot / spell.class['tickcount']) * percentBonus;

	-- Phase 1: Has mana to chain-cast
	local manaUsageSecond = (manaCost / castTime) - manaRegenCasting;
	local secondsPhase1 = math.min(seconds, startMana / manaUsageSecond);
	local secsBetweenSameTarget = math.max(castTime, numtargets * castTime);
	local ticsPerCast = secsBetweenSameTarget / spell.class['interval'];
	ticsPerCast = math.min(ticsPerCast, HealPointsSpells:getNumberOfTicks(spell));
	local spellsCast = secondsPhase1 / castTime;

	local totalHealed = spellsCast * (avgBurst + ticsPerCast * healTick);

	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
		local cycleTime = math.max(castTime, manaCost / manaRegenCasting); -- Seconds between each casting
		secsBetweenSameTarget = math.max(cycleTime, numtargets * cycleTime);
		ticsPerCast = secsBetweenSameTarget / spell.class['interval'];
		ticsPerCast = math.min(ticsPerCast, HealPointsSpells:getNumberOfTicks(spell));
		spellsCast = secondsPhase2 / cycleTime;

		local healedPhase2 = spellsCast * (avgBurst + ticsPerCast * healTick);
		totalHealed = totalHealed + healedPhase2;
	end

	return totalHealed;
end

local function computeTotalHealingLifebloom(spell, healing, seconds, startMana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal)

	--Get the the percent bonus to healing spells (e.g. 1.07 = 7%)
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactorHot, healFactorBurst = HealPointsSpells:getHealingFactor(spell);
	spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);
	
	local avgBurst = (
			spell['burst'] 
			--+ (healing + HealPointsBS:GetBonus('AVG_BURST_LIFEBL')) * healFactorBurst
		) * percentBonus;
		
	local healTick = (
		spell['tick'] 
		--+ (healing + HealPointsBS:GetBonus('AVG_HOT_LIFEBL')) * healFactorHot / spell.class['tickcount']
		) * percentBonus;
	avgBurst = avgBurst * (1 + spellCrit / 200);
	local duration = HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	local numtargets = HealPoints.db.char.hot['numtargets'];

	if (_G["HealPointsOptionsConfigLifeBl"]:GetChecked() == 1) then -- stack lifeblooms
		-- Phase 1: Has enough mana/regen to keep stacking lifeblooms up on numtargets constantly
		numtargets = math.min(3, numtargets); -- max 3 targets
		local cycleTime = duration;
		local manaUsageCycle = numtargets * manaCost;
		local manaRegainCycle = cycleTime * manaRegenCasting;

		local numCycles = seconds / cycleTime;
		if (manaRegainCycle < manaUsageCycle) then -- if you regen more than you use, you'll never go oom.
			numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
		end
		local secondsPhase1 = numCycles * cycleTime;
		local numTicks = numCycles * numtargets * 3 * (HealPointsSpells:getNumberOfTicks(spell) - 1); -- Overwrites before the last tick
		if (numCycles > 2) then -- Ticks lost due to building stacks
			numTicks = numTicks - numtargets * 3 * (HealPointsSpells:getNumberOfTicks(spell) - 1) - 3 * 0.75 * (numtargets * numtargets - numtargets);
		elseif (numCycles == 2) then
			numTicks = numTicks - numtargets * 3 * (HealPointsSpells:getNumberOfTicks(spell) - 1) - 2 * 0.75 * (numtargets * numtargets - numtargets);
		elseif (numCycles == 1) then
			numTicks = numTicks - numtargets * 2 * (HealPointsSpells:getNumberOfTicks(spell) - 1) - 0.75 * (numtargets * numtargets - numtargets);
		end
		local totalHealed = numTicks * healTick;
		if (numCycles > 1) then
			totalHealed = totalHealed + numtargets * (avgBurst + healTick); -- Last tick + burst when oom
		end

		-- Phase 2: Has no mana left, no stacking
		local secondsPhase2 = seconds - secondsPhase1;
		if (secondsPhase2 > 0) then
			cycleTime = manaUsageCycle / manaRegenCasting; --how long it takes to get enough mana to cast a cycle
			numCycles = secondsPhase2 / cycleTime;
			local healedPhase2 = numCycles * numtargets * (avgBurst + healTick * HealPointsSpells:getNumberOfTicks(spell));
			totalHealed = totalHealed + healedPhase2;
		end
		return totalHealed;
	
	else
		-- Phase 1: Has enough mana/regen to keep lifebloom up on numtargets constantly
		local cycleTime = math.max(duration + 1, numtargets * getGCD(spellHaste)); -- time between casts on the same target
		local manaUsageCycle = numtargets * manaCost;
		local regenTimeCasting = math.min(cycleTime, 5 + (numtargets - 1) * getGCD(spellHaste));
		local regenTimeNormal = math.max(0, cycleTime - regenTimeCasting);
		local manaRegainCycle = regenTimeCasting * (manaRegenCasting / 5) + regenTimeNormal * (manaRegenNormal / 5);
		local healingPerCycle = numtargets * (avgBurst + healTick * HealPointsSpells:getNumberOfTicks(spell));
		local numCycles = seconds / cycleTime;
		if (manaRegainCycle < manaUsageCycle) then -- if you regen more than you use, you'll never go oom.
			numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
		end
		local secondsPhase1 = numCycles * cycleTime;
		local totalHealed = numCycles * healingPerCycle;

		-- Phase 2: Has no mana left
		local secondsPhase2 = seconds - secondsPhase1;
		if (secondsPhase2 > 0) then
			local manaToRegen = manaUsageCycle - regenTimeCasting * (manaRegenCasting / 5);
			cycleTime = regenTimeCasting + manaToRegen / (manaRegenNormal / 5);
			numCycles = secondsPhase2 / cycleTime;
			local healedPhase2 = numCycles * healingPerCycle;
			totalHealed = totalHealed + healedPhase2;
		end

		return totalHealed;
	end
end

local function computeTotalHealingHot(spell, healing, seconds, startMana, spelllHaste, manaRegenCasting)
	local hpCast = getHPCast(spell, healing, 0); -- Can't crit
	local manaCost = HealPointsSpells:getManaCost(spell, 0); -- Can't crit
	local duration = HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	local numtargets = HealPoints.db.char.hot['numtargets'];
	local cycleTime = math.max(duration + 1, numtargets * getGCD(spellHaste)); -- time between casts on the same target
	local healingPerCycle = numtargets * hpCast; --total amount healed per cycle
	local manaUsageCycle = numtargets * manaCost; --mana lost per cycle
	local manaRegainCycle = cycleTime * manaRegenCasting; --mana regenerated per cycle

	--local regenTimeCasting = math.min(cycleTime, 5 + (numtargets - 1) * getGCD(spellHaste));
	--local regenTimeNormal = math.max(0, cycleTime - regenTimeCasting);

	-- Phase 1: Has enough mana/regen to keep the HoT up on numtargets constantly
	local numCycles = seconds / cycleTime;
	-- if you regen more than you use, you'll never go oom.
	if (manaRegainCycle < manaUsageCycle) then 
		numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
	end
	local secondsPhase1 = numCycles * cycleTime;
	local totalHealed = numCycles * healingPerCycle;

	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
		cycleTime = max(cycleTime, manaUsageCycle/manaRegenCasting);
		numCycles = secondsPhase2 / cycleTime;
		local healedPhase2 = numCycles * healingPerCycle;
		totalHealed = totalHealed + healedPhase2;
	end

	return totalHealed;
end

--[[
	Caluclates one of the two terms in the HealPoints equation (HealPoints = PowerPoints+EndurancePoints
	Calculates how much healing can be done by an entire mana bar; takes mana regen into account
	@param maxMana: number - The player's current percent [0..100] of their mana
--]]
local function computePowerPoints(healing, spellCrit, spellHaste, maxMana, manaRegenCasting, manaRegenNormal)

	assert(healing, "computePowerPoints - healing is nil");
	assert(spellCrit, "computePowerPoints - spellCrit is nil");
	assert(spellHaste, "computePowerPoints - spellHaste is nil");
	assert(maxMana, "computePowerPoints - maxMana is nil");
	assert(manaRegenCasting, "computePowerPoints - manaRegenCasting is nil");
	assert(manaRegenNormal, "computePowerPoints - manaRegenNormal is nil");

	local mana = HealPoints.db.char.power['mana'] * maxMana / 100;

	if (HealPoints.db.char.power['auto'] == true) then
		local powerfulSpell;
		local maxPowerPoints = 0;
		for i = 1, table.getn(HealPointsSpells.SPELLTABLE) do
			for j = 1, table.getn(HealPointsSpells.SPELLTABLE[i]), 1 do
				local spell = HealPointsSpells.SPELLTABLE[i][j];

				--HealPoints:Print("computerPowerPoints - spell="..Toolkit:VarAsString(spell.class.abbr));

				if (spell ~= nil and HealPointsSpells:getHighestSpellRank(spell.class.classID) == spell['rank'] and spell.class['type'] ~= 'instant') then
					local powerPoints;
					if (spell.class['type'] == 'hot') then
						powerPoints = computeTotalHealingHot(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellHaste, manaRegenCasting, manaRegenNormal);
					elseif (spell.class['type'] == 'regrowth') then
						powerPoints = computeTotalHealingRegrowth(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
					elseif (spell.class['type'] == 'lifebloom') then
						powerPoints = computeTotalHealingLifebloom(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
					else
						powerPoints = computeTotalHealingNormal(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
					end
					if (powerPoints > maxPowerPoints) then
						powerfulSpell = spell;
						maxPowerPoints = powerPoints;
					end
				end
			end
		end
		return maxPowerPoints, powerfulSpell;

	else
		local powerfulSpell = HealPointsSpells:getSpell(HealPoints.db.char.power['spell']);
		local powerPoints;
		if (powerfulSpell.class['type'] == 'hot') then
			powerPoints = computeTotalHealingHot(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellHaste, manaRegenCasting, manaRegenNormal);
		elseif (powerfulSpell.class['type'] == 'regrowth') then
			powerPoints = computeTotalHealingRegrowth(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
		elseif (powerfulSpell.class['type'] == 'lifebloom') then
			powerPoints = computeTotalHealingLifebloom(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
		else
			powerPoints = computeTotalHealingNormal(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
		end
		return powerPoints, powerfulSpell;
	end
end

--[[
	Caluclates one of the two terms in the HealPoints equation (HealPoints = PowerPoints+EndurancePoints
	Calculates how much healing can be done by an entire mana bar; takes mana regen into account
	@param maxMana: percent - The player's current mana (0..100), typically 0%
--]]
local function computeEndurancePoints(healing, spellCrit, spellHaste, maxMana, manaRegenCasting, manaRegenNormal)

	local mana = HealPoints.db.char.endurance['mana'] * maxMana / 100;
	if (HealPoints.db.char.endurance['auto'] == true) then
		local efficientSpell;
		local maxEndurancePoints = 0;
		for i = 1, table.getn(HealPointsSpells.SPELLTABLE), 1 do
			for j = 1, table.getn(HealPointsSpells.SPELLTABLE[i]), 1 do
				local spell = HealPointsSpells.SPELLTABLE[i][j];
				if (spell ~= nil and HealPointsSpells:getHighestSpellRank(spell.class.classID) == spell['rank'] and spell.class['type'] ~= 'instant') then
					local endurancePoints;
					if (spell.class['type'] == 'hot') then
						endurancePoints = computeTotalHealingHot(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellHaste, manaRegenCasting, manaRegenNormal);
					elseif (spell.class['type'] == 'regrowth') then
						endurancePoints = computeTotalHealingRegrowth(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
					elseif (spell.class['type'] == 'lifebloom') then
						endurancePoints = computeTotalHealingLifebloom(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
					else
						endurancePoints = computeTotalHealingNormal(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
					end
					if (endurancePoints > maxEndurancePoints) then
						efficientSpell = spell;
						maxEndurancePoints = endurancePoints;
					end
				end
			end
		end
		return maxEndurancePoints, efficientSpell;

	else
		local endurancePoints = 0;

		local efficientSpell = HealPointsSpells:getSpell(HealPoints.db.char.endurance['spell']);
		if (efficientSpell) then
			if (efficientSpell.class['type'] == 'hot') then
				endurancePoints = computeTotalHealingHot(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellHaste, manaRegenCasting, manaRegenNormal);
			elseif (efficientSpell.class['type'] == 'regrowth') then
				endurancePoints = computeTotalHealingRegrowth(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
			elseif (efficientSpell.class['type'] == 'lifebloom') then
				endurancePoints = computeTotalHealingLifebloom(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
			else
				endurancePoints = computeTotalHealingNormal(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, manaRegenCasting, manaRegenNormal);
			end
		end;
		return endurancePoints, efficientSpell;
	end
end

--[[
	Calculates the player's
		healPoints, powerPoints, endurancePoints
	as well as the most powerfulSpell and the most efficient spell
	@param maxMana: number - The players current mana as a percent [0..100]. 
--]]
local function computeHealpoints(healing, spellCrit, spellHaste, maxMana, manaRegenCasting, manaRegenNormal)
	
	assert(healing, "computePowerPoints - healing is nil");
	assert(spellCrit, "computePowerPoints - spellCrit is nil");
	assert(spellHaste, "computePowerPoints - spellHaste is nil");
	assert(maxMana, "computePowerPoints - maxMana is nil");
	assert(manaRegenCasting, "computePowerPoints - manaRegenCasting is nil");
	assert(manaRegenNormal, "computePowerPoints - manaRegenNormal is nil");

	local powerPoints, powerfulSpell = computePowerPoints(healing, spellCrit, spellHaste, maxMana, manaRegenCasting, manaRegenNormal);
	local endurancePoints, efficientSpell = computeEndurancePoints(healing, spellCrit, spellHaste, maxMana, manaRegenCasting, manaRegenNormal);

	local healPoints = powerPoints + endurancePoints;

	return healPoints, powerPoints, endurancePoints, powerfulSpell, efficientSpell;
end

--[[Compute current and new values of Mana Regeneration
	Returns
		manaRegenNormal
		manaRegenCasting
		manaRegenNormalNew
		manaRegenCastingNew
		percentRegen
--]]
local function getCurrentAndNewManaRegen_Depricated(itemIntDelta, itemSpiDelta, itemCastingRegenDelta, itemManaDelta)
	-- Base stats  
	local _, intellect, _, _ = UnitStat("player", 4);
	local _, spirit, _, _ = UnitStat("player", 5);

	local spiritRegen = StatLogic:GetNormalManaRegenFromSpi(spirit + itemSpiDelta, intellect + itemIntDelta);
	
	--Mana regeneration from Spirit while in combat increased by xxx percent 
	--	e.g. Paladin Holy specialization Meditation "Allows 50% of your mana regeneration from Spirit to continue while in combat"
	local percentRegen = StatLogic:GetNormalManaRegenFromSpi(spirit, intellect);
	
		local castingRegen = HealPointsUtil:GetBaseCombatManaRegen() --base combat regen amount that everyone always gets
			+ itemCastingRegenDelta; --items can no longer have casting regen (removed from game). But i can leave param here
  
	local manaRegenNormal = castingRegen + spiritRegen;
	local manaRegenCasting = castingRegen + percentRegen*spiritRegen;
  
	-- Calculator bonuses
	local calcInt = _G["HealPointsVariables1InputBox"]:GetNumber() * (1 + BonusPercentInt);
	local calcSpi = _G["HealPointsVariables2InputBox"]:GetNumber() * (1 + BonusPercentSpi);
	local calcMana = _G["HealPointsVariables3InputBox"]:GetNumber();
	local calcCastingRegen = _G["HealPointsVariables4InputBox"]:GetNumber();
	local calcNormalRegen = _G["HealPointsVariables5InputBox"]:GetNumber();
	local calcPercentRegen = _G["HealPointsVariables6InputBox"]:GetNumber() / 100;

	local spiritRegenNew = StatLogic:GetNormalManaRegenFromSpi(spirit + itemSpiDelta + calcSpi, intellect + itemIntDelta + calcInt);
	
	local manaRegenNormalNew = castingRegen + spiritRegenNew;
	local manaRegenCastingNew = castingRegen + percentRegen*spiritRegenNew;

	if (_G["HealPointsOptionsConfigReplenish"]:GetChecked() == 1) then -- add replenishment
		local currentMana = UnitManaMax("player") + itemManaDelta + (15 * itemIntDelta);
		local replenishmp5 = 0.0125 * currentMana; -- 0.25% every sec = 1.25% every 5 sec
		manaRegenNormal = manaRegenNormal + replenishmp5;
		manaRegenCasting = manaRegenCasting + replenishmp5;

		local newMana = UnitManaMax("player") + itemManaDelta + (15 * itemIntDelta) + calcMana + 15 * calcInt;
		local replenishmp5New = 0.0125 * newMana; -- 0.25% every sec = 1.25% every 5 sec
		manaRegenNormalNew = manaRegenNormalNew + replenishmp5New;
		manaRegenCastingNew = manaRegenCastingNew + replenishmp5New;
	end

	if (_G["HealPointsOptionsConfigDivinePlea"]:GetChecked() == 1) then -- add divine plea
		local currentMana = UnitManaMax("player") + itemManaDelta + (15 * itemIntDelta);
		local replenishmp5 = 0.0208333 * currentMana; -- 25% every minute = 2.083% every 5 sec
		manaRegenNormal = manaRegenNormal + replenishmp5;
		manaRegenCasting = manaRegenCasting + replenishmp5;

		local newMana = UnitManaMax("player") + itemManaDelta + (15 * itemIntDelta) + calcMana + 15 * calcInt;
		local replenishmp5New = 0.0208333 * newMana; -- 25% every minute = 2.083% every 5 sec
		manaRegenNormalNew = manaRegenNormalNew + replenishmp5New;
		manaRegenCastingNew = manaRegenCastingNew + replenishmp5New;
	end

	return manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew, percentRegen;
end

local function getCurrentAndNewMaxMana_depricated(itemManaDelta, itemIntDelta)
	local currentMana = UnitManaMax("player") + itemManaDelta + (15 * itemIntDelta);

	local intDelta = _G["HealPointsVariables1InputBox"]:GetNumber() * (1 + BonusPercentInt);
	local newMana = currentMana + (_G["HealPointsVariables3InputBox"]:GetNumber() + 15 * intDelta);
	return currentMana, newMana;
end

local function getCurrentAndNewHealing_depricated(itemHealDelta, itemIntDelta, itemSpiDelta, itemAgiDelta)
	local intBonus = 0;
	local spiBonus = 0;
	local agiBonus = 0;
  
	local _, className = UnitClass("player");
	if (className == "PALADIN") then
		intBonus = 0.04 * HealPointsTalents:getTalentRank(HealPointsTalents.PALADIN.HOLYGU); -- 4/8/12/16/20% of intellect
	elseif (className == "PRIEST") then
		spiBonus = 0.05 * HealPointsTalents:getTalentRank(HealPointsTalents.PRIEST.SPIGUI); -- 5/10/15/20/25% of spirit
	elseif (className == "DRUID") then
		intBonus = 0.04 * HealPointsTalents:getTalentRank(HealPointsTalents.DRUID.LUNGUI); -- 4/8/12% of intellect
		agiBonus = 0.35 * HealPointsTalents:getTalentRank(HealPointsTalents.DRUID.NURINS); -- 35/70% of agility
		if (HealPointsUtil:isPlayerBuffUp(HealPointsLoc.BUFF_TREELI)) then
		spiBonus = 0.05 * HealPointsTalents:getTalentRank(HealPointsTalents.DRUID.IMPTOL); -- 5/10% of spirit while in Tree Form
		end 
	elseif (className == "SHAMAN") then
		intBonus = 0.05 * HealPointsTalents:getTalentRank(HealPointsTalents.SHAMAN.NATBLE); -- 5/10/15% of intellect
	end
	intBonus = intBonus; --+ 0.01 * HealPointsBS:GetBonus('HEALFROMINT');

	local healingDelta = intBonus * itemIntDelta + spiBonus * itemSpiDelta + agiBonus * itemAgiDelta;

	local healingDeltaNew = _G["HealPointsVariables7InputBox"]:GetNumber();
	itemIntDelta = itemIntDelta + _G["HealPointsVariables1InputBox"]:GetNumber() * (1 + BonusPercentInt); 
	itemSpiDelta = itemSpiDelta + _G["HealPointsVariables2InputBox"]:GetNumber() * (1 + BonusPercentSpi);
	healingDeltaNew = healingDeltaNew + intBonus * itemIntDelta + spiBonus * itemSpiDelta + agiBonus * itemAgiDelta;

	local currentHealing = GetSpellBonusHealing() + itemHealDelta;

	return currentHealing + healingDelta, currentHealing + healingDeltaNew;
end

local function getCurrentAndNewSpellCrit_depricated(itemIntDelta, itemCritRatingDelta)
	local spellCrit = 0;

	local _, className = UnitClass("player");
	if (className == "PALADIN" or className == "PRIEST") then
		spellCrit = GetSpellCritChance(2); -- Holy
	elseif (className == "DRUID" or className == "SHAMAN") then
		spellCrit = GetSpellCritChance(4); -- Nature
	end

	local critFromDeltaRating = StatLogic:GetEffectFromRating(itemCritRatingDelta, CR_CRIT_SPELL);
	local critFromDeltaInt = StatLogic:GetSpellCritFromInt(itemIntDelta);
	local currentSpellCrit = spellCrit + critFromDeltaRating + critFromDeltaInt;

	local critRatingDelta = itemCritRatingDelta + _G["HealPointsVariables8InputBox"]:GetNumber();
	local intDelta = itemIntDelta + _G["HealPointsVariables1InputBox"]:GetNumber() * (1 + BonusPercentInt);

	critFromDeltaRating = StatLogic:GetEffectFromRating(critRatingDelta, CR_CRIT_SPELL);
	critFromDeltaInt = StatLogic:GetSpellCritFromInt(intDelta);
	local newSpellCrit = spellCrit + critFromDeltaRating + critFromDeltaInt;

	return currentSpellCrit, newSpellCrit;
end

local function getCurrentAndNewSpellHaste_depricated(itemHasteRatingDelta)
	local spellHaste = GetCombatRatingBonus(CR_HASTE_SPELL);

	local hasteFromDelta = StatLogic:GetEffectFromRating(itemHasteRatingDelta, CR_HASTE_SPELL);
	local currentSpellHaste = spellHaste + hasteFromDelta;

	hasteFromDelta = StatLogic:GetEffectFromRating(itemHasteRatingDelta + _G["HealPointsVariables9InputBox"]:GetNumber(), CR_HASTE_SPELL);
	local newSpellHaste = spellHaste + hasteFromDelta;
  
	return currentSpellHaste, newSpellHaste;
end

local function displaySpellStats_depricated(spell) -- Show derived stats about specified spell
	if (BonusPercentInt == nil) then
		self:talentsChanged();
	end

	local _, manaRegenCasting, _, manaRegenCastingNew = getCurrentAndNewManaRegen(0, 0, 0, 0);
	local currentMana, newMana = getCurrentAndNewMaxMana(0, 0);
	local currentHealing, newHealing = getCurrentAndNewHealing(0, 0, 0, 0);
	local currentSpellCrit, newSpellCrit = getCurrentAndNewSpellCrit(0, 0);
	local currentSpellHaste, newSpellHaste = getCurrentAndNewSpellHaste(0);

	local oldHPSec = getHPSec(spell, currentHealing, currentSpellCrit, currentSpellHaste);
	local newHPSec = getHPSec(spell, newHealing, newSpellCrit, newSpellHaste);
	local oldHPMana = getHPMana(spell, currentHealing, currentSpellCrit);
	local newHPMana = getHPMana(spell, newHealing, newSpellCrit);
	local oldHPCast = getHPCast(spell, currentHealing, currentSpellCrit);
	local newHPCast = getHPCast(spell, newHealing, newSpellCrit);
	local oldHPTotalRegen, t1 = getHPTotalRegen(spell, currentHealing, currentSpellCrit, currentSpellHaste, manaRegenCasting, currentMana);
	local newHPTotalRegen, t2 = getHPTotalRegen(spell, newHealing, newSpellCrit, newSpellHaste, manaRegenCastingNew, newMana);
	local oldCrit = currentSpellCrit + HealPointsSpells:getSpellCritBonus(spell);
	local newCrit = newSpellCrit + HealPointsSpells:getSpellCritBonus(spell);
	local oldCastTime = HealPointsSpells:GetCastTime(spell, currentSpellHaste);
	local newCastTime = HealPointsSpells:GetCastTime(spell, newSpellHaste);

--[[
	-- Common stats
	HealPointsCalculatorUI:setSpellStat(1, "HP/s:", oldHPSec, newHPSec, "%d");
	HealPointsCalculatorUI:setSpellStat(2, "HP/mana:", oldHPMana, newHPMana, "%1.1f");
	HealPointsCalculatorUI:setSpellStat(3, "HP/cast:", oldHPCast, newHPCast, "%d");
	HealPointsCalculatorUI:setSpellStat(4, "HP/manabar:", oldHPTotalRegen, newHPTotalRegen, "%d");

	-- Spell specific stats
	if (spell['class'].type ~= 'hot') then
		HealPointsCalculatorUI:setSpellStat(5, "Spell crit(%):", oldCrit, newCrit, "%1.2f");
	else
		local oldHPTick = getHPTick(spell, currentHealing);
		local newHPTick = getHPTick(spell, newHealing);
		HealPointsCalculatorUI:setSpellStat(5, "HP/tick:", oldHPTick, newHPTick, "%d");
	end

	if (spell['class'].type ~= 'hot' and spell['class'].type ~= 'instant' and spell['class'].type ~= 'lifebloom' and
		spell.class.classID ~= HealPointsSpells.SPELL_HOLYNOVA and spell.class.classID ~= HealPointsSpells.SPELL_COH) then
		HealPointsCalculatorUI:setSpellStat(6, "Cast time:", oldCastTime, newCastTime, "%1.2f");
	elseif (spell['class'].type == 'lifebloom') then
		local oldHPTick = getHPTick(spell, currentHealing);
		local newHPTick = getHPTick(spell, newHealing);
		HealPointsCalculatorUI:setSpellStat(6, "HP/tick:", oldHPTick, newHPTick, "%d");
	else
		HealPointsCalculatorUI:setSpellStat(6, nil);
	end

	if (spell['class'].type == 'regrowth') then
		local oldHPTick = getHPTick(spell, currentHealing);
		local newHPTick = getHPTick(spell, newHealing);
		HealPointsCalculatorUI:setSpellStat(7, "HP/tick:", oldHPTick, newHPTick, "%d");
	else
		HealPointsCalculatorUI:setSpellStat(7, nil);
	end

	local healFactor1, healFactor2 = HealPointsSpells:getHealingFactor(spell);
	local spellBonus = HealPointsSpells:getHealingPercentBonus(spell);
	healFactor1 = healFactor1 * spellBonus;
	if (healFactor2 ~= nil) then
		healFactor2 = healFactor2 * spellBonus;
	end
	if (spell['class'].type == 'aoe') then
		healFactor1 = healFactor1 * math.min(5, HealPoints.db.char.hot['numtargets']);
	end
	HealPointsCalculatorUI:setHealFactor(spell, healFactor1, healFactor2);

	HealPointsCalculatorUI:setTimeToOOM(t1, t2);
--]]
end

--[[
	Calculate the HealPoints gained per point of attributes
			intellect
			spirit
			spell power
			crit rating
			haste rating
	returns
		intellectValue
		spiritValue
		spellPowerValue
		critRatingValue
		hasteRatingValue
--]]
function HealPointsCalculator:ComputeHealPointsDelta(currentStats)
	assert(currentStats, "HealPointsCalculator:ComputeHealPointsDelta - currentStats cannot be nil")

	local iv, sv, spv, scrv, shrv;

	local newStats;

	--1 intellect
	local changes = { intellect = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	iv = (newStats.healPoints - currentStats.healPoints);

	--1 spirit
	local changes = { spirit = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	sv = (newStats.healPoints - currentStats.healPoints);

	--1 spell power
	local changes = { spellPower = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	spv = (newStats.healPoints - currentStats.healPoints);

	--1 spell crit rating
	local changes = { spellCritRating = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	scrv = (newStats.healPoints - currentStats.healPoints);

	--1 spell haste rating
	local changes = { spellHasteRating = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	shrv = (newStats.healPoints - currentStats.healPoints);

	return iv, sv, spv, scrv, shrv, currentStats.powerfulSpell, currentStats.efficientSpell;
end


--[[
	Get a table of the player's current (important) stats:
	
			data.playerLevel  		e.g. 85
			data.playerClass  		e.g. "Paladin"
			data.strength
			data.agility
			data.stamina
			data.intellect
			data.spirit
			data.mana				Player's maximum mana
			data.baseManaRegen 		Player's "base" mana regeneration (i.e. tooltip "Combat Regen")

			data.spellPower			Players tooltip "Spell Power" e.g. 3138
			data.spellCritChance	Player's tooltip "Crit Chance" e.g. 16.04%
			data.spellCritRating 	Player's spell "Crit Rating" e.g. Crit rating 263 (+4.36% crit chance) 
			data.spellHaste			Player's spell "Haste" (e.g. Haste +11.31%)
			data.spellHasteRating	Player's spell "Haste Rating" e.g. Haste rating 487 (+11.31% Haste)
			
			data.attackPower		Player's attack power, e.g. 399 (Increases damage with melee weapons by 28.5 damage per second.)
--]]
function HealPointsCalculator:GetCurrentPlayerStats(data)
	--assert(data)
	
	local result = data;
	if (result == nil) then
		result = {};
	end;

	local unit = "player"
	
	result.playerLevel = UnitLevel(unit)   --WoW API
--	result.playerHealth = UnitHealthMax(unit)   --WoW API
	result.playerClass = select(2, UnitClass(unit))	--   WoW API
	
	--Attributes
	--local stat, effectiveStat, posBuff, negBuff = UnitStat(unit, 1); --1=Strength
	result.strength =  UnitStat(unit, 1); --1=Strength   WoW API
	result.agility =   UnitStat(unit, 2); --2=Agility   WoW API
	result.stamina =   UnitStat(unit, 3); --3=Stamina   WoW API
	result.intellect = UnitStat(unit, 4); --4=Intellect   WoW API
	result.spirit =    UnitStat(unit, 5); --5=Spirit   WoW API
	
	-- Mana
	result.mana = UnitManaMax(unit); --WoW API
	result.baseMana = HealPointsUtil:GetBaseMana();

	local manaRegen, combatRegen = GetManaRegen(); --while not in combat, while in combat
	result.normalManaRegen = manaRegen;
	result.baseManaRegen = HealPointsUtil:GetBaseCombatManaRegen();
	result.combatManaRegen = combatRegen;
	result.combatManaRegenExtra = 0;
	result.manaCostModifier = 1.0;

	--Divine Plea option
	if (HealPoints.db.char.divineplea) then
		--Divine Plea: You gain 12% of your total mana over 9 seconds. 2 minute cooldown
		HealPoints:Print("HealPointsCalculator:GetCurrentPlayerStats - Adding Divine Plea");
		if (result.playerClass == "PALADIN") then
			local manaRegenFromDivinePlea = result.mana*0.12/120;
			result.combatManaRegenExtra = result.combatManaRegenExtra + manaRegenFromDivinePlea;
		end;
	end;
	
	--Replenishment option
	if (HealPoints.db.char.replenishment) then
		--Replenishment: Regain 1% of maximum mana per 10 seconds over 15 seconds
		--15 ticks, each one being (mana*1% / 10)
		--or (mana * 1% / 10) per second
		HealPoints:Print("HealPointsCalculator:GetCurrentPlayerStats - Adding Replenishment");
		local manaFromReplenishment = result.mana*0.01/10;
		result.combatManaRegenExtra = result.combatManaRegenExtra + manaFromReplenishment;
	end;

	result.combatManaRegen = result.combatManaRegen + result.combatManaRegenExtra;
	
	-- Spell
	result.spellPower = GetSpellBonusHealing(); --WoW API GetSpellBonusHealing (e.g. 3138)
	result.spellCritRating = GetCombatRating(CR_CRIT_SPELL); --WoW API GetCombatRating(CR_CRIT_SPELL) = 263 ("Crit rating 263 (+4.36% crit chance)")
	--result.spellCritChanceBonus = GetCombatRatingBonus(CR_CRIT_SPELL); --WoW API GetCombatRatingBonus(CR_CRIT_SPELL) = 4.36 ("Crit rating 263 (+4.36% crit chance)")
	result.spellCritChance = GetSpellCritChance(2); --2=Holy  WoW API GetSpellCritChance(2) = 16.038 ("Crit Chance 16.04%")
	
	result.spellHasteRating = GetCombatRating(CR_HASTE_SPELL); --WoW API GetCombatRating(CR_HASTE_SPELL) = 487 ("Haste rating 487 (+11.31% haste)")
	local spellHasteBonus = GetCombatRatingBonus(CR_HASTE_SPELL);  --WoW API GetCombatRatingBonus(CR_HASTE_SPELL) = 11.3108 ("Haste rating 487 (+11.31% haste)")
	result.spellHaste = spellHasteBonus; --haste might come from other sources

	-- Melee
	result.attackPower = UnitAttackPower("unit"); --WoW API e.g. UnitAttackPower("player") = 399
	
	--[[
	-- Mana regen (casting, normal, %casting)
	local manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew, percentRegen = getCurrentAndNewManaRegen(0, 0, 0, 0);
	local manaRegenCastingDelta = manaRegenCastingNew - manaRegenCasting - _G["HealPointsVariables4InputBox"]:GetNumber();
	local manaRegenNormalDelta = manaRegenNormalNew - manaRegenNormal - _G["HealPointsVariables5InputBox"]:GetNumber();
	HealPointsCalculatorUI:setVariableStatInfo(4, manaRegenCasting, manaRegenCastingDelta, "%5.0f");
	HealPointsCalculatorUI:setVariableStatInfo(5, manaRegenNormal, manaRegenNormalDelta, "%5.0f");
	HealPointsCalculatorUI:setVariableStatInfo(6, 100 * percentRegen, 0, "%5.0f");
	--]]

	-- Healing spell power
	--local currentHealing, newHealing = getCurrentAndNewHealing(0, 0, 0, 0);
	--HealPointsCalculatorUI:setVariableStatInfo(7, currentHealing, newHealing - currentHealing - _G["HealPointsVariables7InputBox"):GetNumber(), "%5.0f"];

	-- Spell crit rating
	--HealPointsCalculatorUI:setVariableStatInfo(8, GetCombatRating(11), 0, "%5.0f");

	-- Spell haste rating
	--HealPointsCalculatorUI:setVariableStatInfo(9, GetCombatRating(20), 0, "%5.0f");
	
	return result;
end;		


--[[
	Calculate result = playerStats+changes
		Changes can come from either
			a tooltip link -> HealPointsTooltips:GetChangesTable -> ApplyChanges
			the calculator -> HealPointsCalculatorUI:GetChanges -> ApplyChanges
		@param playerStats:table - a table of the player's stats to which the changes will be applied
		@param changes:table - a table of changes that will be applied to playerStats
	@returns: table - a player stats table containing the new player stats
--]]	
function HealPointsCalculator:ApplyChanges(playerStats, changes)
	assert(playerStats, "HealPointsCalculator:ApplyChanges - playerStats is nil");
	assert(changes, "HealPointsCalculator:ApplyChanges - changes is nil");

	--Get everything source has
	local results = Toolkit:CopyTable(playerStats);
	
	--[[
		data.playerLevel  		e.g. 85
		data.playerClass  		e.g. "Paladin"
		
		data.strength
		data.agility
		data.stamina
		data.intellect
		data.spirit
		
		data.mana				Player's maximum mana
		data.combatManaRegen 		Player's "base" mana regeneration (i.e. tooltip "Combat Regen")

		data.spellPower			Players tooltip "Spell Power" e.g. 3138
		data.spellCritRating 	Player's spell "Crit Rating" e.g. Crit rating 263 (+4.36% crit chance) 
		data.spellCritChance	Player's tooltip "Crit Chance" e.g. 16.04%
		data.spellHaste			Player's spell "Haste" (e.g. Haste +11.31%)
		data.spellHasteRating	Player's spell "Haste Rating" e.g. Haste rating 487 (+11.31% Haste)
			
		data.attackPower		Player's attack power, e.g. 399 (Increases damage with melee weapons by 28.5 damage per second.)
	--]]
	
	if (changes.strength ~= nil) then
		results.strength = playerStats.strength + (changes.strength or 0);
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_STR")*changes.strength;
		
		changes.strength = nil; --remove item
	end;
	
	changes.attackPower = (changes.attackPower or 0);
	if (changes.attackPower ~= nil) then
		results.attackPower = playerStats.attackPower + changes.attackPower;
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_AP")*changes.attackPower;

		changes.attackPower = nil; --remove item
	end;
	
	if (changes.agility ~= nil) then
		results.agility = playerStats.agility + changes.agility;
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_AGI")*changes.agility;

		changes.agility = nil; --remove item
	end;

	--results.stamina = playerStats.stamina + (changes.stamina or 0);

	local recalculateManaRegen = false;

	if (changes.intellect ~= nil) then
		results.intellect = playerStats.intellect + changes.intellect;

		changes.mana = (changes.mana or 0) + HealPointsUtil:GetManaFromInt(changes.intellect);
		changes.spellPower = (changes.spellPower or 0) + HealPointsUtil:GetSpellPowerFromInt(changes.intellect);
		changes.spellCritChance = (changes.spellCritChance or 0) + StatLogic:GetSpellCritFromInt(changes.intellect);

		recalculateManaRegen = true; --we changed our intellect, we have to recompute mana regen
		changes.intellect = nil; --remove item
	end;

	if (changes.spirit ~= nil) then
		results.spirit = playerStats.spirit + changes.spirit;

		recalculateManaRegen = true; --we changed our spirit, we have to recompute mana regen

		changes.spirit = nil; --remove item
	end;
	
	--local modMana = StatLogic:GetStatMod("MOD_MANA");
	--if (modMana ~= 0) then
	--end;
	
	if (changes.mana ~= nil) then
		results.mana = playerStats.mana + changes.mana;
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_MANA")*changes.mana;
		--changes.combatManaRegen = (changes.combatManaRegen or 0) + StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA")*changes.mana;
		--changes.combatManaRegen = (changes.combatManaRegen or 0)+ StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN")*changes.mana;

		recalculateManaRegen = true;
		changes.mana = nil;
	end;

	--HealPoints:Print("results.mana: "..results.mana);

	--MOD_MANA

	if (recalculateManaRegen) then
		local baseManaRegen = HealPointsUtil:GetBaseCombatManaRegen(); --per second
		local manaRegenThroughSpirit = StatLogic:GetNormalManaRegenFromSpi(results.spirit, results.intellect)/5;
		
		local manaRegenBefore = playerStats.normalManaRegen;
		local manaRegenAfter = baseManaRegen + manaRegenThroughSpirit;

		--HealPoints:Print("manaRegenBefore: "..manaRegenBefore);
		--HealPoints:Print("manaRegenAfter: "..manaRegenAfter);

		changes.normalManaRegen = (manaRegenAfter - manaRegenBefore);

		--HealPoints:Debug("baseManaRegen: "..baseManaRegen);


		local combatRegenBefore = playerStats.combatManaRegen;
		local combatRegenAfter = baseManaRegen + StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN")*manaRegenThroughSpirit + playerStats.combatManaRegenExtra;
		changes.combatManaRegen = (combatRegenAfter - combatRegenBefore);
	end;

	
	if (changes.combatManaRegen ~= nil) then
		results.combatManaRegen = playerStats.combatManaRegen + changes.combatManaRegen;
		changes.combatManaRegen = nil;
	--	HealPoints:Print("results.combatManaRegin: "..results.combatManaRegen);
	end;

	if (changes.normalManaRegen ~= nil) then	
		results.normalManaRegen = playerStats.normalManaRegen + changes.normalManaRegen; 
		changes.normalManaRegen = nil;
		--HealPoints:Print("results.normalManaRegen: "..results.normalManaRegen);
	end;

	if (changes.combatManaRegenPercent ~= nil) then
--		results.combatManaRegenPercent = playerStats.combatManaRegenPercent + changes.combatManaRegenPercent;
		changes.combatManaRegenPercent = nil;
	end;

	if (changes.manaCost ~= nil) then
		results.manaCostModifier = playerStats.manaCostModifier + changes.manaCost;
		changes.manaCost = nil;
	end;
	
	--HealPoints:Print("changes.spellPower: "..changes.spellPower);
	if (changes.spellPower ~= nil) then
		results.spellPower = playerStats.spellPower + changes.spellPower;
		changes.spellPower = nil;
	end;

	if (changes.spellCritRating ~= nil) then
		results.spellCritRating = playerStats.spellCritRating + changes.spellCritRating;
		changes.spellCritChance = (changes.spellCritChance or 0) + StatLogic:GetEffectFromRating(changes.spellCritRating, CR_CRIT_SPELL)/100.0;

		changes.spellCritRating = nil;
	end;

	if (changes.spellCritChance ~= nil) then
		results.spellCritChance = playerStats.spellCritChance + (changes.spellCritChance or 0);

		changes.spellCritChance = nil;
	end;

	if (changes.spellHasteRating ~= nil) then
		results.spellHasteRating = (playerStats.spellHasteRating or 0) + changes.spellHasteRating;

		changes.spellHasteRating = nil;

		results.spellHaste = StatLogic:GetEffectFromRating(results.spellHasteRating, CR_HASTE_SPELL);
	end;

	--Remove items from changes table that i know i can't use
	--results.MASTERY_RATING = nil;
	results.SPELL_DMG = nil;
	results.itemType = nil;
	results.link = nil;
	results.diffItemType1 = nil;
	results.diffItemType2 = nil;
	results.EXPERTISE_RATING = nil;
	results.ARMOR = nil;
	results.MELEE_CRIT_RATING = nil;
	results.MELEE_HASTE_RATING = nil;
	results.MELEE_HIT_RATING = nil;
	results.SPELL_HIT_RATING = nil;
	results.EMPTY_SOCKET_META = nil;
	results.EMPTY_SOCKET_RED = nil;
	results.EMPTY_SOCKET_YELLOW = nil;
	results.EMPTY_SOCKET_BLUE = nil;
	results.MAX_DAMAGE = nil;
	results.DPS = nil;
	results.DODGE_RATING = nil;
	results.PARRY_RATING = nil;
	results.ARMOR_BONUS = nil;
	results.SHADOW_RES = nil;
	results.FIRE_RES = nil;
	results.FROST_RES = nil;

	--Remove items from changes table that i know i can't use
	changes.MASTERY_RATING = nil;
	changes.SPELL_DMG = nil;
	changes.itemType = nil;
	changes.link = nil;
	changes.diffItemType1 = nil;
	changes.diffItemType2 = nil;
	changes.EXPERTISE_RATING = nil;
	changes.ARMOR = nil;
	changes.MELEE_CRIT_RATING = nil;
	changes.MELEE_HASTE_RATING = nil;
	changes.MELEE_HIT_RATING = nil;
	changes.SPELL_HIT_RATING = nil;
	changes.EMPTY_SOCKET_META = nil;
	changes.EMPTY_SOCKET_RED = nil;
	changes.EMPTY_SOCKET_YELLOW = nil;
	changes.EMPTY_SOCKET_BLUE = nil;
	changes.MAX_DAMAGE = nil;
	changes.DPS = nil;
	changes.DODGE_RATING = nil;
	changes.PARRY_RATING = nil;
	changes.ARMOR_BONUS = nil;
	changes.SHADOW_RES = nil;
	changes.FIRE_RES = nil;
	changes.FROST_RES = nil;


	--HealPoints:Print("changes.length"..Toolkit:VarAsString(table.maxn(changes)));
	if next(changes) ~= nil then
		HealPoints:Print("HealPointsCalculator:ApplyChanges - changes leftover: "..Toolkit:VarAsString(changes));
	end;

	return results;
end;


--[[
	Based on the stats contained in passed playerStats, calculate the 
			HealPoints, PowerPoints and EndurancePoints, and other stats
		and write them to the same supplied playerStats object:
			- healPoints
			- powerPoints
			- endurancePoints
			- hpSec - (HP/second) Health Points healed per second by the selected spell
			- hpMana - (HP/mana) Health Points healed per point of mana
			- hpCast - (HP/cast) Health Points healed per cast of the selected spell
--]]
function HealPointsCalculator:CalculateHealPoints(playerStats, forceSpell)
	assert(playerStats, "HealPointsCalculator:CalculateHealPoints: Argument 'playerStats' is nil");
	--assert(forceSpell, "HealPointsCalculator:CalculateHealPoints - Argument 'spell' is nil")
		
	HealPoints:Debug("CalculateHealPoints(forceSpell="..Toolkit:VarAsString(forceSpell)..")");
	
	--i need the spell in order to determine the points; the hp/sec, hp/cast, etc all come from the spell.
	local manaRegenNormal = playerStats.normalManaRegen;
	local manaRegenCasting = playerStats.combatManaRegen;

	local currentMana = playerStats.mana;
	local currentHealing = playerStats.spellPower;
	local currentSpellCrit = playerStats.spellCritChance;
	local currentSpellHaste = playerStats.spellHaste;

	--HealPoints:Print("HealPointsCalculator:CalculateHealPoints");

	assert(currentHealing, "CalculateHealPoints - currentHealing is nil");
	assert(currentSpellCrit, "CalculateHealPoints - currentSpellCrit is nil");
	assert(currentSpellHaste, "CalculateHealPoints - currentSpellHaste is nil");
	assert(currentMana, "CalculateHealPoints - currentMana is nil");
	assert(manaRegenCasting, "CalculateHealPoints - manaRegenCasting is nil");
	assert(manaRegenNormal, "CalculateHealPoints - manaRegenNormal is nil");
	
	local healPoints, powerPoints, endurancePoints, powerfulSpell, efficientSpell =
			computeHealpoints(currentHealing, currentSpellCrit, currentSpellHaste, currentMana, manaRegenCasting, manaRegenNormal, forceSpell);

	playerStats.powerfulSpell = powerfulSpell;
	playerStats.efficientSpell = efficientSpell;
	
	self.HP_LastPowerfulCurrent = powerfulSpell;
--	self.HP_LastPowerfulNew = powerfulNew;
	self.HP_LastEfficientCurrent = efficientSpell;
--	self.HP_LastEfficientNew = efficientNew;

	--HealPoints:Print("CalculateHealPoints - powerfulSpell="..Toolkit:VarAsString(powerfulSpell));
	--HealPoints:Print("CalculateHealPoints - efficientSpell="..Toolkit:VarAsString(efficientSpell));

--	if (HealPoints_CalcFrame:IsVisible()) then
		--HealPointsCalculatorUI:setHealPoints(healedPowerful, healedPowerfulNew, healedEfficient, healedEfficientNew);
	--end
	--if (CharacterFrame:IsVisible()) then
--		HealPointsCharUI:setStats(healPointsCurrent, healedPowerful, healedEfficient, powerfulCurrent, efficientCurrent);
--	end
	
	if (forceSpell) then
		playerStats.hpSec = getHPSec(forceSpell, currentHealing, currentSpellCrit, currentSpellHaste);
		playerStats.hpMana = getHPMana(forceSpell, currentHealing, currentSpellCrit);
		playerStats.hpCast = getHPCast(forceSpell, currentHealing, currentSpellCrit);
		playerStats.hpManaBar = getHPTotalRegen(forceSpell, currentHealing, currentSpellCrit, currentSpellHaste, manaRegenCasting, currentMana);
	end;

	playerStats.powerPoints = powerPoints;
	playerStats.endurancePoints = endurancePoints;
	playerStats.healPoints = healPoints;
end;

--[[
	Returns an object representing the players current stats and healpoint calculation results
	Note: This object might be calculated on the fly, or might be the cached result of a recent calculation
--]]
local _cachedCurrentHealPoints = nil;

function HealPointsCalculator:GetCurrentHealPoints(forceSpell)
	HealPoints:Debug("GetCurrentHealPoints(forceSpell="..Toolkit:VarAsString(forceSpell)..")");

	if (forceSpell == nil) then
		if (_cachedCurrentHealPoints) then
			return _cachedCurrentHealPoints;		
		end;
	end;

	local currentStats = HealPointsCalculator:GetCurrentPlayerStats();
	HealPointsCalculator:CalculateHealPoints(currentStats, forceSpell);

	if (forceSpell == nil) then
		_cachedCurrentHealPoints = currentStats;
	end;

	return currentStats;
end;

---Called by the controller HealPoints:ClearCache() when any cached stats are now invalid
function HealPointsCalculator:ClearCache()
	_cachedCurrentHealPoints = nil;
end;